/* tkXwin.c - some modest X11 extensions to Tk... */

#include <tk.h>
#include <X11/Xatom.h>

#ifndef HAVE_SHARED_XSS
/*
 * some systems have only libXss.a (no libXss.so)
 *
 * so, everything under !HAVE_SHARED_XSS is taken from lib/Xss/XScrnSaver.c
 */

#define NEED_EVENTS
#define NEED_REPLIES
#include <X11/extensions/Xext.h>
#include <X11/Xlibint.h>
#include <X11/Xutil.h>
#include <X11/extensions/extutil.h>
#include <X11/extensions/saverproto.h>
#endif

#include <X11/extensions/scrnsaver.h>
#include <stdio.h>
#include <stdlib.h>

#include <tkInt.h>

/*
   tkXwin::idletime ?-displayof window?
 */

#ifndef HAVE_SHARED_XSS
static XExtensionInfo _screen_saver_info_data;
static XExtensionInfo *screen_saver_info = &_screen_saver_info_data;
static /* const */ char *screen_saver_extension_name = ScreenSaverName;

#define ScreenSaverCheckExtension(dpy,i,val) \
  XextCheckExtension (dpy, i, screen_saver_extension_name, val)
#define ScreenSaverSimpleCheckExtension(dpy,i) \
  XextSimpleCheckExtension (dpy, i, screen_saver_extension_name)

static int close_display();
static Bool wire_to_event();
static Status event_to_wire();
static /* const */ XExtensionHooks screen_saver_extension_hooks = {
    NULL,                               /* create_gc */
    NULL,                               /* copy_gc */
    NULL,                               /* flush_gc */
    NULL,                               /* free_gc */
    NULL,                               /* create_font */
    NULL,                               /* free_font */
    close_display,                      /* close_display */
    wire_to_event,                      /* wire_to_event */
    event_to_wire,                      /* event_to_wire */
    NULL,                               /* error */
    NULL,                               /* error_string */
};

static XEXT_GENERATE_FIND_DISPLAY (find_display, screen_saver_info,
                                   screen_saver_extension_name, 
                                   &screen_saver_extension_hooks,
                                   ScreenSaverNumberEvents, NULL)

static XEXT_GENERATE_CLOSE_DISPLAY (close_display, screen_saver_info)


static Bool wire_to_event (dpy, re, event)
    Display *dpy;
    XEvent  *re;
    xEvent  *event;
{
    XExtDisplayInfo *info = find_display (dpy);
    XScreenSaverNotifyEvent     *se;
    xScreenSaverNotifyEvent     *sevent;

    ScreenSaverCheckExtension (dpy, info, False);

    switch ((event->u.u.type & 0x7f) - info->codes->first_event) {
    case ScreenSaverNotify:
        se = (XScreenSaverNotifyEvent *) re;
        sevent = (xScreenSaverNotifyEvent *) event;
        se->type = sevent->type & 0x7f;
        se->serial = _XSetLastRequestRead(dpy,(xGenericReply *) event);
        se->send_event = (sevent->type & 0x80) != 0;
        se->display = dpy;
        se->window = sevent->window;
        se->window = sevent->root;
        se->state = sevent->state;
        se->kind = sevent->kind;
        se->forced = True;
        if (sevent->forced == xFalse)
            se->forced = False;
        se->time = sevent->timestamp;
        return True;
    }
    return False;
}

static Status event_to_wire (dpy, re, event)
    Display *dpy;
    XEvent  *re;
    xEvent  *event;
{
    XExtDisplayInfo *info = find_display (dpy);
    XScreenSaverNotifyEvent     *se;
    xScreenSaverNotifyEvent     *sevent;

    ScreenSaverCheckExtension (dpy, info, 0);

    switch ((re->type & 0x7f) - info->codes->first_event) {
    case ScreenSaverNotify:
        se = (XScreenSaverNotifyEvent *) re;
        sevent = (xScreenSaverNotifyEvent *) event;
        sevent->type = se->type | (se->send_event ? 0x80 : 0);
        sevent->sequenceNumber = se->serial & 0xffff;
        sevent->root = se->root;
        sevent->window = se->window;
        sevent->state = se->state;
        sevent->kind = se->kind;
        sevent->forced = xFalse;
        if (se->forced == True)
            sevent->forced = xTrue;
        sevent->timestamp = se->time;
        return 1;
    }
    return 0;
}
#endif


static int
tkXwin_idletime (ClientData clientData,
                 Tcl_Interp *interp,
                 int         objc,
                 Tcl_Obj    *CONST objv[]) {
    int              ok,
                     skip;
    Tk_Window        tkWin;
#ifdef  HAVE_SHARED_XSS
    XScreenSaverInfo ssi;
#else
    Display                   *dpy;
    xScreenSaverQueryInfoReply rep;
    xScreenSaverQueryInfoReq  *req;
    XExtDisplayInfo           *info;
#endif

    tkWin = (Tk_Window) clientData;
    if ((skip = TkGetDisplayOf (interp, objc - 1, objv + 1, &tkWin)) < 0)
       return TCL_ERROR;
    if (objc - skip != 1) {
        Tcl_WrongNumArgs (interp, 1, objv, "?-displayof window?");
        return TCL_ERROR;
    }

#ifdef  HAVE_SHARED_XSS
    if (!XScreenSaverQueryInfo (Tk_Display (tkWin),
				DefaultRootWindow (Tk_Display (tkWin)), &ssi))
	ssi.idle = 0;

    Tcl_SetLongObj (Tcl_GetObjResult (interp), ssi.idle);
    return TCL_OK;
#else
    dpy = Tk_Display (tkWin);
    info = find_display (dpy);

    Tcl_SetLongObj (Tcl_GetObjResult (interp), 0L);
    ScreenSaverCheckExtension (dpy, info, TCL_OK);

    LockDisplay (dpy);

    GetReq (ScreenSaverQueryInfo, req);
    req -> reqType = info -> codes -> major_opcode;
    req -> saverReqType = X_ScreenSaverQueryInfo;
    req -> drawable = DefaultRootWindow (dpy);
    ok = _XReply (dpy, (xReply *) &rep, 0, xTrue);

    UnlockDisplay (dpy);
    SyncHandle ();

    Tcl_SetLongObj (Tcl_GetObjResult (interp), ok ? rep.idle : 0L);
    return TCL_OK;
#endif
}


/*
   tkXwin::changeprop window property type format mode data
 */

static int
tkXwin_changeprop (ClientData clientData,
                  Tcl_Interp *interp,
                  int         objc,
                  Tcl_Obj    *CONST objv[]) {
    int           i,
                  mode,
                  nelem,
                  result;
    unsigned long format,
                 *data;
    Atom          prop,
                  type;
    Tcl_Obj      *elemPtr;
    Tk_Window     tkWin;
    Window        window;

    /* same order as in <X11/X.h> */
    static char *modeStrings[] = {
        "replace",       "prepend",       "append",
        NULL
    };

    if (objc != 7) {
        Tcl_WrongNumArgs (interp, 1, objv,
                          "window property type format mode data");
        return TCL_ERROR;
    }

    tkWin = (Tk_Window) clientData;
    if (Tcl_GetLongFromObj (interp, objv[1], &window) != TCL_OK) {
        if ((tkWin = Tk_NameToWindow (interp, Tcl_GetStringFromObj (objv[1],
                                                                    NULL),
                                      tkWin)) == NULL)
            return TCL_ERROR;
        Tk_MakeWindowExist (tkWin);
        window = Tk_WindowId (tkWin);
        fprintf (stderr, "window id is 0x%x\n",(unsigned long) window);
        if (Tk_IsMapped (tkWin))
            fprintf (stderr, "window is aleady mapped\n");
    }

    prop = Tk_InternAtom (tkWin, Tcl_GetStringFromObj (objv[2], NULL));

    type = Tk_InternAtom (tkWin, Tcl_GetStringFromObj (objv[3], NULL));

    if (Tcl_GetLongFromObj (interp, objv[4], &format) != TCL_OK)
        return TCL_ERROR;
    if (format != 32) {
        Tcl_SetResult (interp, "format must be 32", TCL_STATIC);
        return TCL_ERROR;
    }

    if (Tcl_GetIndexFromObj(interp, objv[5], modeStrings, "mode", 0,
                            &mode) != TCL_OK)
        return TCL_ERROR;

    if (Tcl_ListObjLength (interp, objv[6], &nelem) != TCL_OK)
        return TCL_ERROR;

    data = (unsigned long *)
                        ckalloc ((unsigned) ((nelem + 1) * sizeof *data));
    for (i = 0; i < nelem; i++)
        if ((Tcl_ListObjIndex (interp, objv[6], i, &elemPtr) != TCL_OK)
                || (Tcl_GetLongFromObj (interp, elemPtr, &data[i])
                        != TCL_OK)) {
            ckfree ((char *) data);
            return TCL_ERROR;
        }

    Tcl_SetIntObj (Tcl_GetObjResult (interp),
                   XChangeProperty (Tk_Display (tkWin), window,
                                    prop, type, format, mode,
                                    (unsigned char *) data, nelem));

    ckfree ((char *) data);
    return TCL_OK;
}


int
Tkxwin_Init (Tcl_Interp *interp) {
    if (Tcl_InitStubs (interp, "8.0", 0) == NULL)
        return TCL_ERROR;
    if (Tk_InitStubs (interp, "8.0", 0) == NULL)
        return TCL_ERROR;

    if (Tcl_PkgRequire (interp, "Tk", TCL_VERSION, 0) == NULL)
        return TCL_ERROR;

    if (Tcl_PkgProvide (interp, "tkXwin", PACKAGE_VERSION) != TCL_OK)
        return TCL_ERROR;

    Tcl_CreateObjCommand (interp, "tkXwin::changeprop", tkXwin_changeprop,
                          (ClientData) Tk_MainWindow (interp),
                          (Tcl_CmdDeleteProc *) NULL);

    Tcl_CreateObjCommand (interp, "tkXwin::idletime", tkXwin_idletime,
                          (ClientData) Tk_MainWindow (interp),
                          (Tcl_CmdDeleteProc *) NULL);

    return TCL_OK;
}
